/*
    Copyright (C) 2005, 2006  Vansteenhuyse Hubert <hvansteenhuyse@yahoo.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*******************************************************************************
    LOGIC LAYER
    -----------
        General application classes.

                                                            Sunday, 4 dec 2005.
*******************************************************************************/
#include "CHexEd.h"
#include <string.h>
#include <unistd.h>
#ifdef __UNIX__
    #include <ncurses.h>
#endif
unsigned long lGoto = 0;
int nGotoPct = 0;

CHexEd::CHexEd(){
    memset(this, 0, sizeof(CHexEd));
    mScreenID = SCR_INTRO;
    mEditMode = EDT_VIEW;
    min = new CStr();
    mSearch = new CStr();
    mBuffer = 0;
    mScreen = new CScreen();
    mInput = new CUserInput();
    mx = 1;
    my = 0;
    mSW_hexside = 0;
    mTop = 0;
    mCursor = 0;
    mDir = new ioDIR();
}

CHexEd::~CHexEd(){
    delete min;
    delete mSearch;
    if(mBuffer) delete mBuffer;
    if(mDir)delete mDir;
    delete mScreen;
    delete mInput;
}

int CHexEd::LoadFile(char *path){
    if(mBuffer)delete mBuffer;
    mBuffer = new CBuffer(path);
    mScreenID = SCR_BUFFER;
    mTop = 0;
    my = 0;
    mx = 0;
    mCursor = 0;
    return mBuffer->err();
}

void CHexEd::save(){
    long AdresXY = 0;
    if(mBuffer && mEditMode == EDT_EDIT && strlen(input) > 0){
            //log("input = '%s'", input);
            AdresXY = GetBufferIndexXY();
            //long CBuffer::write(unsigned long fileaddress, char *buf, long amount){
            mBuffer->WriteFile(mTop + AdresXY, mBuffer->GetBuffer() + AdresXY, 1);
            memset(input, 0, 3);
            if(mBuffer->err()) mScreen->ShowError(mBuffer->err(), "Failed to read file %s", mBuffer->GetPath());
    }
}


void CHexEd::HighLight(int sw_hi, int x, int y){
    char ch, hex[3];
    long AdresXY, MaxAdresXY;
    int MinAsciiX, HexDumpX, AsciiX;

    if(!mBuffer) return;
        AdresXY = GetBufferIndexXY(x, y);
        MaxAdresXY = mBuffer->GetBufferTail() - mBuffer->GetBuffer();
        if(AdresXY > MaxAdresXY) AdresXY = MaxAdresXY;
        if(AdresXY > mBuffer->GetAmountLastLoaded() || mBuffer->GetAmountLastLoaded() == 0) return;
        ch = mBuffer->GetBuffer()[AdresXY];
        atoh(ch, hex);
        hex[2] = 0;

        MinAsciiX = mScreen->GetMinAsciiDump();
        if(x < MinAsciiX){
            HexDumpX = x - (x % 1);
            AsciiX = MinAsciiX + (HexDumpX / 3);
        }else{
            HexDumpX = ((x - MinAsciiX) * 3) + 1;
            AsciiX = x;
        }


        //highlight...
        if(x < MinAsciiX) mScreen->setDisplayModus((!sw_hi ? DISP_NORMAL : (mEditMode == EDT_EDIT ? DISP_HIGHLIGHTED : DISP_NORMAL))); else mScreen->setDisplayModus((!sw_hi ? DISP_NORMAL : DISP_REVERSE));
        mScreen->printxy(HexDumpX, y, hex);
        if(x >= MinAsciiX) mScreen->setDisplayModus((!sw_hi ? DISP_NORMAL : (mEditMode == EDT_EDIT ? DISP_HIGHLIGHTED : DISP_NORMAL)));  else mScreen->setDisplayModus((!sw_hi ? DISP_NORMAL : DISP_REVERSE));
        ch = mScreen->GetDisplayChar(ch);
        mScreen->printcharxy(AsciiX, y, ch);
        mScreen->gotoxy(mx, my);
}

void CHexEd::DrawFooter(){
    unsigned char ch;
    unsigned long ulFSize;
    float pct;
    long AdresXY = 0, maxx, lenfname;
    //unsigned int maxx = mScreen->GetMaxX();
    char buf[2048];
        if(mBuffer){
            memset(buf, 0, 2048);
            ulFSize = mBuffer->GetFileSize();
            pct = (((float)mCursor + (mCursor > 0 ? 1 : 0)) / (float)(ulFSize > 0 ? ulFSize : 1)) * 100;
            AdresXY = GetBufferIndexXY();
            ch = mBuffer->GetBuffer()[AdresXY];
            sprintf(buf + strlen(buf), "%s", (SW_DISKREAD == 1 ? "R" : " "));
            if(SW_READ_ONLY == 1) sprintf(buf + strlen(buf), "--RO--");
            else if(mEditMode == EDT_EDIT)  sprintf(buf + strlen(buf), "--OVERWRITE--");
            sprintf(buf + strlen(buf), " [%.2f%%%%]", (float)pct);
            sprintf(buf + strlen(buf), " [%lu/%lu]", mCursor, ulFSize);
            sprintf(buf + strlen(buf), " [%.3d,0x%.2X,%s] [", ch,ch, AsciiXlat(ch));

            AdresXY = 0;
            maxx = mScreen->GetMaxX();
            lenfname = strlen(mBuffer->GetPath());
            while(AdresXY < lenfname && (long)(strlen(buf) + strlen(mBuffer->GetPath() + AdresXY)) > maxx){
                if(!AdresXY) sprintf(buf + strlen(buf), "...");
                AdresXY++;
            }
            sprintf(buf + strlen(buf), "%s]", mBuffer->GetPath() + AdresXY );
        }else
            sprintf(buf, " [BUFFER EMPTY]. Basic Commands <:h> <:ls> <:q>    +   <enter>");
    mScreen->DrawCommandLine(buf);
    mScreen->gotoxy(mx, my);
}

//computes the index inside mBuffer buffer where my + mx are pointing to
long CHexEd::GetBufferIndexXY(int x, int y){
    long res = 0;
    if(!mBuffer) return res;
    if(x == -1 && y == -1){
        x = mx;
        y = my;
    }
    res = mScreen->GetHexDumpLineSize() * y;
    if(x < mScreen->GetMinAsciiDump())
        res += (x - (x % 2)) / 3;
    else
        res += x - mScreen->GetMinAsciiDump();
    return res;
}


/*
    movedDirection:
        = 0 = no move
        > 0 = down
        < 0 = up
*/
void CHexEd::SetCursor(int x, int y){
    int movedDirection = my - y;
    unsigned long CursorAddress;
    if(mBuffer){
        CursorAddress = mTop + GetBufferIndexXY(x, y);
        if(CursorAddress > mBuffer->GetFileSize()) CursorAddress = (movedDirection > 0 ? mBuffer->GetFileSize() : 0);
    }else{
        CursorAddress = 0;
        x = 0;
        y = 0;
    }
    mCursor = CursorAddress;
    //log("FileAddress %d, Amount Loaded %d, mCursor %d", mBuffer->GetFileAddress(), mBuffer->GetAmountLoaded(), mCursor);
    HighLight(0, mx, my);
    mx = x;
    my = y;
    HighLight(1, mx, my);
    mScreen->gotoxy(mx, my);
}

/******************************************************************************************************************
 ******************************************************************************************************************
 cursor pointers
 ******************************************************************************************************************
 ******************************************************************************************************************/
/*
    top = first char upper left corner screen
*/
void CHexEd::MoveTop(KEYID ky){
    long    LoadAmount = 0,
            PageSize = mScreen->GetHexDumpScreenSize(),
            LineSize = mScreen->GetHexDumpLineSize();
    unsigned long LoadFromAddress = 0;//, OldTop = mTop;

        if(!mBuffer) return;
        switch(ky){
            case KPGD:
                LoadFromAddress = mTop + PageSize;
                LoadAmount = PageSize;
                break;
            case KDOWN:
                LoadFromAddress = mTop + LineSize;
                LoadAmount = PageSize;
                break;
            case KPGU:
                LoadFromAddress = mTop - PageSize;
                LoadAmount = PageSize * -1;
                if(LoadFromAddress > mBuffer->GetFileSize())LoadFromAddress = 0;
                break;
            case KUP:
                LoadFromAddress = mTop - LineSize;
                LoadAmount = PageSize * -1;
                if(LoadFromAddress > mBuffer->GetFileSize())LoadFromAddress = 0;
                break;
            case KHOME:
                mTop = 0;
                mCursor = 0;
                LoadFromAddress = mTop;
                LoadAmount = PageSize;
                break;
            case KNUL:
                LoadFromAddress = mTop;
                LoadAmount = PageSize;
                break;
            default:
                return;
                break;
        };

        SW_DISKREAD = 1;
        DrawFooter();
        mScreen->commit();
        mBuffer->ReadFrom(LoadFromAddress, LoadAmount);
        SW_DISKREAD = 0;
        mTop = mBuffer->GetFileAddress() - mBuffer->GetAmountLastLoaded();
        //log("MoveTop: OldTop was %d, new mTop = %d", OldTop, mTop);
}


COMMANDID CHexEd::parseCommand(char *str){
    unsigned long i = 0;
    if(strcmp(str, ":q") == 0) return CMD_QUIT;
    if(str[0] == ':' && str[1] == 'l' && str[2] == 's' && str[3] == 0) return CMD_LS;
    if(str[0] == ':' && str[1] == 'w' && str[2] == 0) return CMD_SET_WRITEMODE;
    if(str[0] == ':' && str[1] == 'r' && str[2] == 0) return CMD_SET_READMODE;
    if(str[0] == ':' && str[1] == 'h' && str[2] == 0) return CMD_HELP;
    if(str[0] == ':' && (str[1] >= '0' && str[1] <= '9')){
        lGoto = 0;
        nGotoPct = 0;
        i = (unsigned long)strlen(str);
        if(str[i - 1] == '%'){
            str[i - 1] = 0;
            nGotoPct = atoi(str + 1);
        }else{
            lGoto = (unsigned long)atol(str + 1);
        }
        return CMD_GOTO;
    }
    return CMD_UNKNOWN;
}

void CHexEd::DrawScreen(SCREENID id){
    int x = mx,
        y = my;
    SCREENID scrid = (id ? id : mScreenID);

    mScreen->cursor_on(0);
    switch(scrid){
        case SCR_INTRO:
            mScreen->DrawIntro();
            DrawScreen(SCR_COMMAND);
            break;
        case SCR_CMD_LS:
            mScreen->DrawLS(mDir);
            break;
        case SCR_COMMAND:
            if(strlen(min->get())){
                mScreen->DrawCommandLine(min->get());
            }else{
                mScreen->gotoxy(mx, my);
            }
            break;
        case SCR_BUFFER:
            if(!mBuffer){
                mScreen->HexDump(0, 0);
                DrawFooter();
            }else{
                //scherm opladen...
                if(mBuffer->GetFileSize() && mBuffer->GetBufferSize() != mScreen->GetHexDumpScreenSize()){
                        mBuffer->ReadFrom(
                            mBuffer->GetFileAddress() - mBuffer->GetAmountLastLoaded(),
                            mScreen->GetHexDumpScreenSize()
                        );
                        /*herberekenen van mTop en mCursor...*/
                        mTop = mBuffer->GetFileAddress() - mBuffer->GetAmountLastLoaded();
                        if(mBuffer->GetFileSize() && mScreen->GetMaxY() - 1 > 0){
                            x = 1;
                            while(mTop + mScreen->GetHexDumpLineSize() * y >= mBuffer->GetFileSize() || y >= mScreen->GetMaxY() - 1 ) y--;
                        }
                    SetCursor(x, y);
                }

                mScreen->HexDump(mBuffer->GetBuffer(), mBuffer->GetBufferTail());
                HighLight(1, mx, my);
                if(mBuffer->err())
                    mScreen->ShowError(mBuffer->err(), "Failed to read file %s", mBuffer->GetPath());
                else
                    DrawFooter();
            }
            break;
        case SCR_CMD_HELP:
            mScreen->DrawHelp();
            break;
        case SCR_CMD_COMMANDLINEHELP:
            mScreen->DrawCommandlineHelp();
            break;
        case SCR_CMD_SEARCH:
            mScreen->DrawSearch(mSearch, mSW_hexside);
            break;
        default:
            mScreen->DrawUnknownScreen(mScreenID);
    };
    mScreen->commit();
    mScreen->cursor_on(1);
}


/*appends the user input in a CStr object, returns an ACTION to be taken*/
ACTION CHexEd::doInput(CStr *appendto){
    ACTION act;
    int ch;
      do{
            ch = mInput->GetInput();

            if(SW_QUIT) return ACT_CANCEL;
            if(ch == KBACK){
                 if(appendto->getsize() > 1) appendto->del();
                 ch = 1;
            }else if(ch == KCANCEL){
                act = ACT_CANCEL;
                break;
            }else if(ch == '\n'){
                act = ACT_PARSECOMMAND;
                break;
            }else if(ch < 20){
                continue;
            }
            if(ch > 1)appendto->append((char)ch);
            DrawScreen(SCR_COMMAND);
      }while(ch);
    return act;
}

void CHexEd::dohelp(){
    mScreenID = SCR_CMD_HELP;
    DrawScreen();
    mInput->GetInput();
    mScreenID = SCR_BUFFER;
    DrawScreen(SCR_COMMAND);
}

/*
    list files
*/
void CHexEd::dolistfiles(){
      CStr *path = new CStr();
      int ch, swDone = 0;
      DIRFILE *f;
      long lenp;

          path->append(mDir->Directory);
          if(mDir->CountElements() == 0){
                mDir->Reset();
                mScreen->DrawCommandLine("LS: Opening folder...");
                mScreen->commit();
                ioLowLevel::ioListDirectory(mDir, path->get(), "*.*");
                mScreen->DrawCommandLine("LS: Sorting...");
                mScreen->commit();
                mDir->Sort();
                path->clear();
                path->append(mDir->Directory);
          }
          mDir->mSearchFor.clear();
          mScreenID = SCR_CMD_LS;
            do{
                DrawScreen();
                ch = mInput->GetInput();
                if(SW_QUIT) return;

                if(ch == KCANCEL) swDone = 1;


                if(ch == '\n'){
                    if(strlen(mDir->mSearchFor.get()) > 0 && mDir->mSearchFor.get()[0] == ':'){
                        if(mDir->mSearchFor.get()[1] == 'c' && mDir->mSearchFor.get()[2] == 'd' ){
                            lenp = 3;
                            while(mDir->mSearchFor.get()[lenp] == ' ') lenp++;
                            chdir(mDir->mSearchFor.get() + lenp);
                            path->clear();
                            path->append(mDir->mSearchFor.get() + lenp);
                            mDir->Reset();
                            mScreen->DrawCommandLine("LS: Opening folder...");
                            mScreen->commit();
                            ioLowLevel::ioListDirectory(mDir, path->get(), "*.*");
                            mScreen->DrawCommandLine("LS: Sorting...");
                            mScreen->commit();
                            mDir->Sort();
                        }
                    }else{
                          f = mDir->ElementAt(mDir->GetSelectedItem());
                          if(f && f->IsFolder){
                              if(strcmp(f->FileName, "..") == 0){
                                  lenp = strlen(path->get());
                                  while(lenp && !swDone){
                                      lenp = strlen(path->get());
                                      if(path->get()[lenp - 1] == ioLowLevel::ioGetFileSeparatorChar())swDone = 1;
                                      path->del();
                                  }
                              }else{
                                  path->append(ioLowLevel::ioGetFileSeparatorChar());
                                  path->append(f->FileName);
                              }
                              mDir->Reset();
                              mScreen->DrawCommandLine("LS: Opening folder...");
                              mScreen->commit();
                              //log("Path: '%s'", path->get());
                              ioLowLevel::ioListDirectory(mDir, path->get(), "*.*");
                              mScreen->DrawCommandLine("LS: Sorting...");
                              mScreen->commit();
                              mDir->Sort();
                              swDone = 0;
                        }else{
                            swDone = 1;
                            path->append(ioLowLevel::ioGetFileSeparatorChar());
                            path->append(f->FileName);
                            LoadFile(path->get());
                        }
                    }
                }else if(ch == KBACK){
                    mDir->mSearchFor.del();
                }else if(ch >= 32  && ch <= 126 && strlen(path->get()) < 1023){
                    mDir->mSearchFor.append(ch);
                    ch = 0;
                }

                mDir->mLastKey = (KEYID)ch;

            }while(!swDone);

        delete path;
        mScreenID = SCR_BUFFER;
        DrawScreen(SCR_COMMAND);
}

int CHexEd::dosearchinput(){
    CStr hx;
    int dosearch = 0, ch, swDone = 0, hxval;
    long x;
    if(!mSearch) return 0;
    mScreenID = SCR_CMD_SEARCH;
    DrawScreen();

    do{
        DrawScreen();
        ch = mInput->GetInput();
            if(SW_QUIT) return 0;
            x = mSearch->getsize() % mScreen->GetHexDumpLineSize();
            if(ch == KCANCEL) swDone = 1;
            if(ch == '\n'){ dosearch = 1; swDone = 1; }
            if(ch == KBACK){ mSearch->del(); hx.clear(); }

            if(ch == KLEFT || ch == KRIGHT){
                mSW_hexside = (mSW_hexside ? 0 : 1);
            }

            if(mSW_hexside){
                if(ch >= '0' && ch <= '9' || LCASE(ch) >= 'a' && LCASE(ch) <= 'f'){
                    hx.append(ch);
                    hxval = htoi(hx.get());
                    if(hx.getsize() == 1){
                        mSearch->append(hxval);
                    }else{
                        mSearch->del();
                        mSearch->append(hxval);
                        hx.clear();
                    }
                }
            }else{
                if(ch == 9 || ch >= 32  && ch <= 126){
                    mSearch->append(ch);
                    ch = 0;
                }
            }
    }while(!swDone);
    mScreenID = SCR_BUFFER;
    if(!SW_QUIT) DrawScreen();
    return dosearch;
}


int CHexEd::search(int sw_fwd){
    char buf[64];
    float pct;
    int swfound = 0, swdone = 0;
    CBuffer         *cbuf = 0;
    long x = mx;
    unsigned long   count,
                    ulFSize,
                    bytes,
                    Address,
                    Top = mCursor,
                    Cursor = mCursor;

    //een nieuwe CBuffer aanmaken en daarin zoeken...
    //Copy operator voorzien in CBuffer ???
    if(!mBuffer || mScreen->GetMaxY() < 1) return swfound;
    cbuf = new CBuffer(mBuffer);
    ulFSize = (cbuf->GetFileSize() == 0 ? 1 : cbuf->GetFileSize());
    bytes = ((unsigned long)mScreen->GetHexDumpScreenSize() > mSearch->getsize() ? mScreen->GetHexDumpScreenSize() : mSearch->getsize()) * 2;
        //forward/backward search
        do{
            pct = ((float)Cursor / (float)(ulFSize)) * 100;
            sprintf(buf, " Searching %.2f%%%%...", (float)pct);
            mScreen->DrawCommandLine(buf);
            mScreen->commit();
            if(sw_fwd){
                cbuf->ReadFrom(Cursor, bytes);
            }else{
                if(Cursor - bytes > ulFSize) bytes = Cursor;
                cbuf->ReadFrom(Cursor - bytes, bytes + 1);
            }
            //log("Search %s: cursor %d, bytes %d", (sw_fwd ? "fwd" : "back"), Cursor, bytes);
            Address = cbuf->GetFileAddress() - (sw_fwd ? 0 : cbuf->GetAmountLastLoaded());
            count = (sw_fwd ? 0 : bytes);
            while(  !swfound
                        &&
                    (sw_fwd ? Cursor + mSearch->getsize() <= Address : Address <= Cursor)
            ){
                if(SW_QUIT) return 0;
                //log("cursor %d, Address %d, char: 0x%X", Cursor, Address, cbuf->GetBuffer()[count]);
                if(memcmp(cbuf->GetBuffer() + count, mSearch->get(), mSearch->getsize()) == 0) swfound = 1;
                if(!swfound){
                    if(sw_fwd){
                        count++;
                        Cursor++;
                    }else{
                        if(!Cursor){
                            Address = 1;    //backwards search: loop verlaten
                        }else{
                            count--;
                            Cursor--;
                        }
                    }
                }
                mInput->DoEvents(); //adem tijdens de loop...
            }
            if(!swfound)Top = Cursor;
            if(sw_fwd){
                if(Cursor + mSearch->getsize() > ulFSize) swdone = 1;
            }else{
                if(Cursor > ulFSize || !Cursor){ swdone = 1; }
            }
        }while(!swfound && !swdone);

    if(cbuf)delete cbuf;

    if(swfound){
        //berekenen van de mCursor en mTop...
        mCursor = Cursor;
        //log("#bytes/screen: %d. Cursor (%d)  -  Top (%d)  = %d", mScreen->GetHexDumpScreenSize(), Cursor, Top, Cursor - Top);
        if(Cursor - mTop >= (unsigned long)mScreen->GetHexDumpScreenSize() ){
            mTop = mCursor - mScreen->GetHexDumpLineSize();
            if(mTop > mBuffer->GetFileSize())mTop = 0;
            mBuffer->ReadFrom(mTop, mScreen->GetHexDumpScreenSize());
        }
        Cursor = mCursor - mTop;
        my = Cursor / (mScreen->GetHexDumpLineSize() == 0 ? 1 : mScreen->GetHexDumpLineSize());
        mx = Cursor % mScreen->GetHexDumpLineSize();
        if(x <= mScreen->GetMinAsciiDump()){
            mx *= 3;
            mx++;
        }else
            mx += mScreen->GetMinAsciiDump();

        DrawScreen();
        //HighLight(0, mx, my);
    }
    if(!swfound)
        mScreen->ShowError(0, " SEARCH PATTERN NOT FOUND...");
    else
        DrawFooter();
    mScreen->commit();
    return swfound;
}


int CHexEd::main(int argc, char *argv[]){
    char *filename;
    long hxval, x, y;
    int         ch = 0, prevch = 0;
    COMMANDID   idcmd;
    char        *errmsg = 0;

    /*****************************************************************************
    OPENEN BESTAND...
    ******************************************************************************/
    if(argc > 1){
        int argCount = 1;
        if(strcmp(argv[argCount], "--help") == 0 || strcmp(argv[argCount], "-h") == 0){
            argCount++;
            mScreenID = SCR_CMD_COMMANDLINEHELP;
            DrawScreen();
            SW_SCREENREADY = 1;
            mInput->GetInput();
            exit(0);
        }else if(strcmp(argv[argCount], "--read-only") == 0 || strcmp(argv[argCount], "-r") == 0){
            argCount++;
            SW_READ_ONLY = 1;
        }
        if(argCount < argc){
            filename = argv[argCount];
            LoadFile(filename);
        }
    }
    DrawScreen();
    SW_SCREENREADY = 1;

    while(!SW_QUIT){
        errmsg = 0;
        ch = mInput->GetInput();
        if(SW_QUIT) return 0;
        /***********************************************************************************
         ** EDT_VIEW of EDT_EDIT
         **********************************************************************************/
        if(mEditMode == EDT_VIEW || mEditMode == EDT_EDIT){
            //ZZ to quit
            if(ch == 'Z' && prevch == 'Z'){ SW_QUIT = 1; continue; }
            prevch = ch;
            if(mEditMode == EDT_VIEW && ch == ':') mEditMode = EDT_COMMAND;

            /***************************************************************
                Scherm: SCR_BUFFER (=hexadecimale dump van een bestand...)
            ***************************************************************/
            if(mBuffer && mScreenID ==SCR_BUFFER && mBuffer->GetFileSize() > 0){
                x = mx;
                y = my;
                if(ch == KHOME){
                    save();
                    MoveTop(KHOME);
                    DrawScreen();
                    SetCursor(1, 0);

                }
                if(ch == KRIGHT || (ch == 'l' && mEditMode == EDT_VIEW)){
                    save();
                    if(x < mScreen->GetMaxHexDump() - 3) x += 3;
                    else if(x >= mScreen->GetMaxHexDump() - 3 && x <= mScreen->GetMaxHexDump()-1) x = mScreen->GetMinAsciiDump();
                    else x += 1;
                    if(mTop + GetBufferIndexXY(x, y) >= mBuffer->GetFileSize()){
                        if(x >= mScreen->GetMinAsciiDump())
                            continue;
                        else
                            x = mScreen->GetMinAsciiDump();
                    }
                    if( x > mScreen->GetMaxX() - 2){
                      x = 1;
                      ch = KDOWN;
                    }else{
                        SetCursor(x, y);
                        DrawFooter();
                        continue;
                    }
                }
                if(ch == KLEFT || (ch == 'h' && mEditMode == EDT_VIEW)){
                    save();
                    if(x < mScreen->GetMaxHexDump()) x -= 3;
                    else if(x == mScreen->GetMinAsciiDump()) x = mScreen->GetMaxHexDump() - 3;
                    else x -= 1;

                    if(x == mScreen->GetMaxHexDump() - 3){
                        while(mTop + GetBufferIndexXY(x, y) >= mBuffer->GetFileSize()) x -= 3;
                    }

                    if(x < 1){
                        x = mScreen->GetMaxX() - 2;
                        ch = KUP;
                    }else{
                        SetCursor(x, y);
                        DrawFooter();
                        continue;
                    }
                }
                if(ch == KPGD){
                    save();
                    MoveTop(KPGD);
                    DrawScreen();
                    while(mTop + GetBufferIndexXY(x, y) > mBuffer->GetFileSize())y--;
                    SetCursor(x, y);
                    DrawFooter();
                    continue;
                }
                if(ch == KPGU){
                    save();
                    if(mTop == 0) continue;
                    MoveTop(KPGU);
                    DrawScreen();
                    SetCursor(x, y);
                    DrawFooter();
                    continue;
                }
                if(ch == KDOWN || (ch == 'j' && mEditMode == EDT_VIEW)){
                    save();
                    if(mTop + GetBufferIndexXY(x, y) + mScreen->GetHexDumpLineSize() >= mBuffer->GetFileSize())continue;
                    if(y + 1 >= mScreen->GetMaxY() - 1){
                        MoveTop(KDOWN);
                        DrawScreen();
                    }else{
                        y++;
                    }
                    SetCursor(x, y);
                    DrawFooter();
                    continue;
                }
                if(ch == KUP || (ch == 'k' && mEditMode == EDT_VIEW)){
                    save();
                    if(y - 1 < 0){
                        if(mTop == 0) continue;
                        MoveTop(KUP);
                        DrawScreen();
                    }else{
                        y--;
                    }
                    SetCursor(x, y);
                    DrawFooter();
                    continue;
                }
            }
        }

        /***********************************************************************************
         ** EDT_VIEW
         **********************************************************************************/
        if(mEditMode == EDT_VIEW){
            if(ch == '/' && mBuffer){
                if(dosearchinput()) search(1);
                continue;
            }

            if(ch == 'n' && mBuffer && mSearch->getsize()){
                if(mCursor + 1 <= mBuffer->GetFileSize()){
                    mCursor++;
                    if(!search(1)) mCursor--;
                    continue;
                }else{
                    mScreen->ShowError(0, "Search past tail of buffer...");
                    mScreen->commit();
                }
            }

            if(ch == 'p' && mBuffer && mSearch->getsize()){
                if(mCursor){
                    mCursor--;
                    if(!search(0)) mCursor++;
                    continue;
                }else{
                    mScreen->ShowError(0, "Search at top of buffer...");
                    mScreen->commit();
                }
            }

            if(ch == 'i' || ch == 'o'){
                if(!mBuffer) continue;
                memset(input, 0, 3);
                if(mBuffer->IsReadOnly()){
                   if(SW_READ_ONLY == 1)
                        mScreen->ShowError(0, "WRITE MODE IS DISABLED (commandline --read-only)!");
                    else
                        mScreen->ShowError(0, "BUFFER IS READ ONLY. Type :w<enter> to enable writing...");
                    mScreen->commit();
                    memset(input, 0, 3);
                }else{
                    mEditMode = EDT_EDIT;
                    DrawFooter();
                    HighLight(1, mx, my);
                }
                continue;
            }

            if(ch == KCANCEL){
                errmsg = 0;
                if(mBuffer)mBuffer->noerr();
                DrawScreen();
            }
        }

        /***********************************************************************************
         ** EDT_EDIT
         **********************************************************************************/
        if(mEditMode == EDT_EDIT){
            if(ch == KCANCEL){
                memset(input, 0, 3);
                mEditMode = EDT_VIEW;
                DrawFooter();
                HighLight(1, mx, my);
                continue;
            }

            if(mx < mScreen->GetMinAsciiDump()){
                if(LCASE(ch) >= 'a' && LCASE(ch) <='f' || ch >='0' && ch <= '9'){
                    if(strlen(input) > 1) memset(input, 0, 3);
                    sprintf(input + strlen(input), "%c", ch);
                    hxval = htoi(input);
                    mBuffer->GetBuffer()[GetBufferIndexXY()] = hxval;
                    HighLight(1, mx, my);
                    DrawFooter();
                }
            }else{
                mBuffer->GetBuffer()[GetBufferIndexXY()] = ch;
                input[0] = ch;
                input[1] = 0;
                HighLight(1, mx, my);
                DrawFooter();
            }
        }

        /***********************************************************************************
         ** EDT_COMMAND
         **********************************************************************************/
        if(mEditMode == EDT_COMMAND){
            prevch = 0;
            min->append((char)ch);
            DrawScreen(SCR_COMMAND);
            switch(doInput(min)){
                case ACT_CANCEL:
                    break;
                case ACT_PARSECOMMAND:
                    idcmd = parseCommand(min->get());
                    mEditMode = EDT_VIEW;
                    switch(idcmd){
                      case CMD_QUIT:
                          SW_QUIT = 1;
                          break;
                      case CMD_LS:
                          dolistfiles();
                          break;
                      case CMD_SET_WRITEMODE:
                          if(SW_READ_ONLY == 1){
                             errmsg = "^ WRITE MODE IS DISABLED (commandline --read-only)!";
                          }else if(mBuffer && mBuffer->IsReadOnly()){
                            mBuffer->SetReadOnly(0);
                            errmsg = "^ Writing is ON, OFF = :r<enter>. ATTENTION: CHANGES ARE APPLIED IMMEDIATELY!";
                          }
                          break;
                      case CMD_SET_READMODE:
                          if(mBuffer && !mBuffer->IsReadOnly()){
                            mBuffer->SetReadOnly(1);
                            errmsg = "^ Writing is OFF, ON = :w<enter>.";
                          }
                          break;
                      case CMD_HELP:
                          dohelp();
                          break;

                      case CMD_GOTO:
                          if(mBuffer){
                              if(nGotoPct){
                                  if(nGotoPct > 100) nGotoPct = 100; else if(nGotoPct < 0) nGotoPct = 0;
                                  if(mBuffer->GetFileSize())
                                      lGoto = mBuffer->GetFileSize() / 100 * nGotoPct;
                                  else
                                      lGoto = 1; //force error...
                              }

                              if(lGoto <= mBuffer->GetFileSize()){
                                  mCursor = lGoto;
                                  mTop = lGoto;
                                  MoveTop(KNUL);
                                  lGoto = mCursor - mTop;
                                  my = lGoto / (mScreen->GetHexDumpLineSize() == 0 ? 1 : mScreen->GetHexDumpLineSize());
                                  mx = lGoto % mScreen->GetHexDumpLineSize();
                                  mx *= 3;
                                  mx++;
                                  SetCursor(mx, my);
                              }else
                                 errmsg = "^specified address past length of file...";
                          }
                          break;

                      default:
                          errmsg = "^ERROR: unknown editor command.";
                          break;
                    }
                    break;
            }
            if(SW_QUIT) return 0;
            min->clear();
            mEditMode = EDT_VIEW;
            mScreenID = SCR_BUFFER;
            DrawScreen();
            DrawScreen(SCR_COMMAND);
            if(errmsg) mScreen->DrawCommandLine(errmsg, 1);
        }
    }
    mScreen->clrscr();
    return 0;
}

void CHexEd::SetFocus(int swGotFocus){
    if(!SW_SCREENREADY) return;
    SW_GOTFOCUS = swGotFocus;
    mScreen->cursor_on(swGotFocus);
}

